/* 
 * Copyright (C) 2006 Registro.br. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistribution of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/* $Id: SheppCommandFunctions.H 685 2006-04-24 13:42:10Z cesar $ */
/** @file SheppCommandFunctions.H
 *  @brief EPP command-line shell client command functions include file
 */
#ifndef __SHEPP_COMMAND_FUNCTIONS_H__
#define __SHEPP_COMMAND_FUNCTIONS_H__

#include "SheppGlobal.H"
#include "SheppStrUtil.H"
#include "SheppObjSet.H"
#include "SheppPrint.H"

#include "IoException.H"
#include "TransportException.H"
#include "EppException.H"
#include "XmlException.H"

#include "Login.H"
#include "Logout.H"

// Functions for eppsh commands
int cmd_login __P((char *));
int cmd_logout __P((char *));
int cmd_hello __P((char *));
int cmd_poll __P((char *));
int cmd_domain __P((char *));
int cmd_contact __P((char *));
int cmd_brorg __P((char *));

int cmd_xmlcmd __P((char *));
int cmd_xmlrsp __P((char *));

int cmd_server __P((char *));
int cmd_port __P((char *));
int cmd_client_pem __P((char *));
int cmd_root_pem __P((char *));
int cmd_pass __P((char *));
int cmd_user __P((char *));
int cmd_login_pw __P((char *));
int cmd_login_new_pw __P((char *));
int cmd_connect __P((char *));

int cmd_help __P((char *));
int cmd_about __P((char *));
int cmd_quit __P((char *));

int cmd_beauty __P((char *));
int cmd_debug __P((char *));

/// Populates set of available commands
void init_commands()
{
  SheppCommand cmd;

  //non-EPP commands
  cmd.reset("?", cmd_help, "This help screen");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("help", cmd_help, "This help screen");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("about", cmd_about, "shepp version information");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("exit", cmd_quit, "Exit shepp\n");
  _commands.insert(_commands.end(), cmd);

  //runtime flags
  cmd.reset("beauty", cmd_beauty, "Turn XML beautifier on/off");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("debug", cmd_debug, "Turn debug messages on/off\n");
  _commands.insert(_commands.end(), cmd);

  //connection related commands
  cmd.reset("server", cmd_server, "Sets server address");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("port", cmd_port, "Sets server port");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("client-pem", cmd_client_pem,
	    "Sets client.pem certificate file location");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("root-pem", cmd_root_pem,
	    "Sets root.pem certificate file location");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("pass", cmd_pass, "Sets SSL certificate passphrase");
  _commands.insert(_commands.end(), cmd);  
  cmd.reset("user", cmd_user, "Sets EPP Login username");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("pw", cmd_login_pw, "Sets EPP Login password");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("newpw", cmd_login_new_pw, "Sets new EPP Login password");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("connect", cmd_connect, "Establishes connection\n");
  _commands.insert(_commands.end(), cmd);

  //EPP commands
  cmd.reset("login", cmd_login, "EPP Login");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("logout", cmd_logout, "EPP Logout");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("hello", cmd_hello, "EPP Hello");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("poll", cmd_poll, "EPP Poll related actions");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("domain", cmd_domain, "EPP Domain related actions");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("contact", cmd_contact, "EPP Contact related actions");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("brorg", cmd_brorg, "EPP BrOrg related actions\n");
  _commands.insert(_commands.end(), cmd);

  //XML related commands
  cmd.reset("xmlcmd", cmd_xmlcmd, "Show last command XML");
  _commands.insert(_commands.end(), cmd);
  cmd.reset("xmlrsp", cmd_xmlrsp, "Show last response XML");
  _commands.insert(_commands.end(), cmd);
}

/// Given a word check if it's a valid SheppCommand
/**
   @param   name of command check validity
   @return pointer to a SheppCommand if found, NULL pointer if not found
*/
const SheppCommand* find_command(string name)
{
  CMD_CONTAINER<SheppCommand>::const_iterator it;
  for (it = _commands.begin(); it != _commands.end(); it++) {
    if ((*it).name == name) {
      return &(*it);
    }
  }

  return ((const SheppCommand *) NULL);
}

/// Sets server address [and port] based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_server(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.size() != 0) {
    string new_port;
    string new_server;
    SheppStrUtil::split(words[0], new_server, new_port, ":", true);
    if (new_server != _server) {
      _server = new_server;
    }
    if (new_port != "" && _port != atoi(new_port.c_str())) {
      _port = atoi(new_port.c_str());
    }
  }

  if (_server != "") {
    printf("server %s:%d\n", _server.c_str(), _port);
  } else {
    printf("no server address given.\n");
  }

  return 0;
}

/// Sets server port based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_port(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.size() == 1) {
    if (_port != atoi(words[0].c_str())) {
      _port = atoi(words[0].c_str());
    }
  }

  printf("port %d\n", _port);

  return 0;
}

/// Sets client.pem certificate file location
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_client_pem(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.empty()) {
    printf("Current client.pem is %s\n", _client_pem.c_str());
    return 0;
  } else if (words.size() > 1) {
    printf("usage: client-pem <file>\n");
    return -1;
  }

  _client_pem = words[0];

  return 0;
}

/// Sets root.pem certificate file location
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_root_pem(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.empty()) {
    printf("Current root.pem is %s\n", _root_pem.c_str());
    return 0;
  } else if (words.size() > 1) {
    printf("usage: root-pem <file>\n");
    return -1;
  }

  _root_pem = words[0];

  return 0;
}

/// Sets SSL certificate passphrase based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_pass(char *arg)
{
  _passphrase = (string) arg;
  return 0;
}

/// Sets login user based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_user(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.size() == 0) {
    if (_user != "") {
      printf("username: %s\n", _user.c_str());
      return 0;
    } else {
      printf("no username given.\n");
      return -1;
    }
  }

  _user = words[0];

  return 0;
}

/// Sets EPP login password based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_login_pw(char *arg)
{
  _login_pw = (string) arg;
  return 0;
}

/// Sets new EPP login password based on input command-line
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_login_new_pw(char *arg)
{
  _login_new_pw = (string) arg;
  _new_pw = true;
  return 0;
}

/// Connects to previously specified server and port
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_connect(char *arg)
{
  if (cmd_server(arg) != 0) {
    return -1;
  }

  bool exception = false;

  // Session code goes here
  try {
    _session = auto_ptr<Session>(new Session(_server, _port));
    _session->connect(_client_pem, _root_pem, _passphrase);
    printf("Connected to %s\n", _server.c_str());
    
    Greeting *greeting = _session->get_greeting();  
    if (_debug) {
      if (greeting) {
	printf("Greeting received\n");
      }
    } //_debug

    _connected = true;

  } catch (const IoException &e) {
    printf("ERROR! IO Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
  } catch (const TransportException &e) {
    printf("ERROR! Transport Exception [%d]:\n[%s]\n[%s]\n",
	   e.get_code(), e.get_msg().c_str(), e.get_low_level_msg().c_str());
    exception = true;
  } catch (const GeneralException &e) {
    printf("ERROR! General Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
  }

  if (exception) {
#if 0
    printf("error connecting to %s:%d\n", _server.c_str(), _port);
#endif
    return -1;
  }

  return 0;
}

/// Turns XML beautifier ON and OFF
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_beauty(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.empty()) {
    if (_beauty) {
      printf("XML beautifier is ON\n");
    } else {
      printf("XML beautifier is OFF\n");
    }
    return 0;
  } else if (words.size() > 1) {
    printf("usage: beauty [on|off]\n");
    return -1;
  }

  if (words[0] == "on") {
    _beauty = true;
    printf("XML beautifier now ON\n");
  } else if (words[0] == "off") {
    _beauty = false;
    printf("XML beautifier now OFF\n");
  } else {
    printf("usage: beauty [on|off]\n");
    return -1;
  }

  return 0;
}

/// Turns debug messages ON and OFF
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_debug(char *arg)
{
  vector<string> words = SheppStrUtil::parse_line(arg);

  if (words.empty()) {
    if (_debug) {
      printf("Debug is ON\n");
    } else {
      printf("Debug is OFF\n");
    }
    return 0;
  } else if (words.size() > 1) {
    printf("usage: beauty [on|off]\n");
    return -1;
  }

  if (words[0] == "on") {
    _debug = true;
    printf("Debug now ON\n");
  } else if (words[0] == "off") {
    _debug = false;
    printf("Debug now OFF\n");
  } else {
    printf("usage: beauty [on|off]\n");
    return -1;
  }

  return 0;
}

/// Shows last XML command
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_xmlcmd(char *arg)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  if (_beauty) {
    // Convert to UTF8
    string last_command("");
    StrUtil::iso88591_to_utf8(_session->get_last_command(), last_command);
    
    if (_session->get_last_command() == "") {
      printf("\n");
      return 0;
    }
    try {
      StrUtil str_util;
      printf("%s\n", str_util.xml_beautifier(last_command).c_str());
    } catch (const XmlException &e) {
      printf("ERROR! XML Exception [%d]:\n[%s]\n", e.get_code(),
	     e.get_msg().c_str());
    }
  } else {
    printf("%s\n", _session->get_last_command().c_str());
  }
  return 0;
}

/// Shows last XML response
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_xmlrsp(char *arg)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  if (_beauty) {
    // No need to convert to UTF8
    string last_response = _session->get_last_response();
    
    if (last_response == "") {
      printf("\n");
      return 0;
    }
    try {
      StrUtil str_util;
      printf("%s\n", str_util.xml_beautifier(last_response).c_str());
    } catch (const XmlException &e) {
      printf("ERROR! XML Exception [%d]:\n[%s]\n", e.get_code(),
	     e.get_msg().c_str());
    }
  } else {
    printf("%s\n", _session->get_last_response().c_str());
  }
  return 0;
}

// EPP action processing functions

/// Prints "command sent ok" message
void print_cmd_sent_ok()
{
  printf("Ok! Use 'xmlcmd' and 'xmlrsp' to view command/response "
	 "XML code.\n");
}

/// Send and EPP Action
/**
   @return 0 if ok, -1 otherwise
*/
int process_action(Action &act)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  bool exception = false;

  try {
    _session->process_action(&act);
  } catch (const EppException &e) {
    printf("ERROR! EPP Exception [%d]:\n[%s]\n", e.get_code(),
	   e.get_msg().c_str());  
    exception = true;
  } catch (const IoException &e) {
    printf("ERROR! IO Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
    _connected = false;
  } catch (const TransportException &e) {
    printf("ERROR! Transport Exception [%d]:\n[%s]\n[%s]\n",
	   e.get_code(), e.get_msg().c_str(), e.get_low_level_msg().c_str());
    exception = true;
    _connected = false;
  } catch (const GeneralException &e) {
    printf("ERROR! General Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
  }

  if (exception) {
    return -1;
  }

  print_cmd_sent_ok();
  return 0;
}

/// Send an EPP Login
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_login(char *arg)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  if (_user == "") {
    printf("no username given.\n");
    return -1;
  }

  Login act;
  LoginCmd *cmd = act.get_command();

  cmd->set_clID(_user);
  cmd->set_pw(_login_pw);

  if (_new_pw) {
    _new_pw = false;
    cmd->set_new_pw(_login_new_pw);
  }

  if (process_action(act) != 0) {
    return -1;
  }  

  return 0;
}

/// Send an EPP Logout
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_logout(char *arg)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  Logout act;

  if (process_action(act) != 0) {
    return -1;
  }  

  return 0;
}

/// Send an EPP HELLO
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_hello(char *arg)
{
  if (!_connected) {
    printf("not connected\n");
    return -1;
  }

  bool exception = false;

  try {
    if (_debug) {
      printf("Sending EPP Hello\n");
    }
    _session->send_hello();
  } catch (const EppException &e) {
    printf("ERROR! EPP Exception [%d]:\n[%s]\n", e.get_code(),
	   e.get_msg().c_str());  
    exception = true;
  } catch (const IoException &e) {
    printf("ERROR! IO Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
    _connected = false;
  } catch (const TransportException &e) {
    printf("ERROR! Transport Exception [%d]:\n[%s]\n[%s]\n",
	   e.get_code(), e.get_msg().c_str(), e.get_low_level_msg().c_str());
    exception = true;
    _connected = false;
  } catch (const GeneralException &e) {
    printf("ERROR! General Exception [%d]:\n[%s]\n", 
	   e.get_code(), e.get_msg().c_str());
    exception = true;
  }

  if (exception) {
    return -1;
  }

  print_cmd_sent_ok();
  return 0;
}

// non-EPP commands

/// Leaves shepp
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_quit(char *arg)
{
  printf("Bye\n");
  exit(0);
}

/// Prints help
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_help(char *arg)
{
  CMD_CONTAINER<SheppCommand>::const_iterator it;
  for (it = _commands.begin(); it != _commands.end(); it++) {
    printf("%-16s %s\n", (*it).name.c_str(), (*it).brief.c_str());
  }

  return 0;
}

/// Prints version information
/**
   @param arg   input command-line
   @return 0 if ok, -1 otherwise
*/
int cmd_about(char *arg)
{
  printf("shepp version " SHEPP_VERSION ", an EPP client shell!\n");
  printf("Copyright 2006 Registro.br <libepp@registro.br>\n");
  printf("shepp is distributed with libepp-nicbr: "
	 "http://registro.br/epp/index-EN.html\n");
#if USE_BR_DOMAINS
  printf("Built with BrDomain EPP extension support.\n");
#endif //USE_BR_DOMAINS

  return 0;
}

// -------------------------------------
// Buggy Auto-complete related functions
// -------------------------------------

// char** shepp_completion __P((const char *, int, int));

// // Attempts to auto-complete "text" with known commands
// char** cpp_completion(string txt)
// {
//   set<string> matches_set;
//   int len = strlen(txt.c_str());

//   set<SheppCommand>::const_iterator it;
//   for (it = commands.begin(); it != commands.end(); it++) {
//     if (strncmp((*it).name.c_str(), txt.c_str(), len) == 0) {
//       matches_set.insert((*it).name);
//     }
//   }

//   char **matches = (char **) NULL;

//   if (matches_set.size() > 0) {
//     const int _DIMENSION = 128;
//     matches = new char*[_DIMENSION];
//     int j;
//       for (j = 0; j < _DIMENSION; j++) {
//         matches[j] = new char[_DIMENSION];
// 	bzero(matches[j], _DIMENSION);
//       }
    
//     set<string>::const_iterator it_str;
//     for (it_str = matches_set.begin(), j = 0;
// 	 it_str != matches_set.end();
// 	 it_str++, j++) {
//       matches[j] = new char[_DIMENSION];
//       strncpy(matches[j], (*it_str).c_str(), strlen((*it_str).c_str()) + 1);
// #if 0
//       printf("found: [%s]\n", matches[j]);
// #endif
//     }
//   }

//   return matches;
// }

// char** shepp_completion (const char *text, int start, int end)
// {
//   char **matches = (char **)NULL;;

//   // if 'text' starts at position '0' then it should be checked
//   // against the command list
//   if (start == 0) {
//     //  matches = rl_completion_matches(text, command_generator);
//     matches = cpp_completion(text);
//   }

//   return (matches);
// }

#endif //__SHEPP_COMMAND_FUNCTIONS_H__
